/*********************************************************************
 *
 *                  BMS_CAN File
 *
 *********************************************************************
 * FileName:        BMS_CAN.c
 * Processor:       PIC18F45K80
 * Compiler:        Microchip C18 v3.41
 * Company:         KIT - CN - IPE
 *
 * Author           Date         Comment
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Ott W.           08.02.2016   Release
 *********************************************************************/

#include "BMS_UI_Main.h"



/*********************************************************************
*
*                             Globals
*
*********************************************************************/
static CAN_MAILBOX gCAN_MAILBOX[11];
static uint8_t     gCAN_dataC[8];
static uint8_t     gCAN_Rx_Overrun;

extern CAN_CONFIG  gCAN_CONFIG;
extern IRC_MSG_BUF gIRC_MSG_BUF;
extern SERIAL_NR_t  gSerial;
extern uint8_t  STATE_STATUS;
static uint8_t     gCAN_Data[8];



/*********************************************************************
 * Function:        void CAN_Init( CAN_CONFIG *config )
 *
 * Overview:        API Function
 *                  Use this function to initialize ECAN module
 *
 * PreCondition:    None
 *
 * Input:           config
 *
 * Output:          return: CAN_OK / CAN_NOK
 *
 * Side Effects:    change CAN_Mailbox
 **********************************************************************/
int8_t CAN_Init( CAN_CONFIG *config )
{
    // Enter CAN module into config mode
    ECANSetOperationMode(ECAN_OP_MODE_CONFIG);

    // Enter CAN module into Mode 1
    ECANSetFunctionalMode(ECAN_MODE_1);

    // Initialize CAN Timing
    // 1 Mbps @ 64Mhz
    // BRP = 2, Propagation Delay =1, Phase Segment 1 = 8, Phase Segment 2 = 6
    // SJW = 1, Multiple bit sampling is on


    BRGCON1 = ((config->SJW-1) << 6) | (config->BRP-1);
    BRGCON2 = (config->PHSEG2_MODE << 7)      | \
              (config->BUS_SAMPLE_MODE << 6)  | \
              ((config->PHSEG1-1) << 3)       | \
              (config->PROPSEG-1);
    BRGCON3 = (config->WAKEUP_MODE << 7)      |
              (config->FILTER_MODE << 6)      |
              (config->PHSEG2-1);
   
    // Setup Programmable buffers
    //  B0 is a receive buffer AND B2,B3,B4,B5 are Transmit buffers
    ECANSetBnTxRxMode(B0, ECAN_BUFFER_RX);

    // Initialize Receive Masks
    //  The first mask is used that accepts all SIDs and no EIDs
    ECANSetRXM0Value(0x0F0F, ECAN_MSG_STD);

    //  The second mask is used for Slave0-Slave15
    ECANSetRXM1Value(0x0F00, ECAN_MSG_STD);

    // Enable Filters
    RXFCON0 = 0;     
    RXFCON1 = 0;
    ECANSetRXF0Value(0x100, ECAN_MSG_STD);  // Startvalue for Master
    ECANSetRXF1Value(0x200, ECAN_MSG_STD);  // Startvalue for Slave0

    // Assign Filters to Masks
    ECANLinkRXF0Thru3ToMask(ECAN_RXM0, ECAN_RXM1, 0xFF, 0xFF);  


    // Assign Filters to Buffers
    ECANLinkRXF0F1ToBuffer(RXB0, RXB1);     //Assign Filter 0 to RXB0, and Filter 1 to RXB1
    ECANLinkRXF2F3ToBuffer(0xFF, 0xFF);     //Assign Filter 2 to B0

    // Enable Port C Pin 2 for CAN capture
    CIOCON = 0x20;           //CANCAP on

    // Enter CAN module into normal mode
    ECANSetOperationMode(ECAN_OP_MODE_NORMAL);

    // Set Receive Mode for buffers
    RXB0CON = 0x00;
    RXB1CON = 0x00;
    B0CON   = 0x00;

    Load_Mailbox( config->SlaveNo );

    return CAN_OK;
}



/*********************************************************************
 * Function:        int8_t CAN_Init_Mailbox( CAN_MAILBOX *mbox)
 *
 * Overview:        API Function
 *                  Use this function to initialize CAN Mailbox
 *
 * PreCondition:    None
 *
 * Input:           *mbox
 *
 * Output:          return: CAN_OK / CAN_NOK
 *
 * Side Effects:    change CAN_Mailbox
 **********************************************************************/
int8_t CAN_Init_Mailbox( CAN_MAILBOX *mbox)
{
  uint8_t   i = mbox->BOX_No;

  gCAN_MAILBOX[i].BOX_No            = i;
  gCAN_MAILBOX[i].CAN_Id            = mbox->CAN_Id;
  gCAN_MAILBOX[i].DLC               = mbox->DLC;
  gCAN_MAILBOX[i].ACCEPTANCE_FILTER = mbox->ACCEPTANCE_FILTER;

  return CAN_OK;
}



/*********************************************************************
 * Function:        int8_t CAN_Write( CAN_MAILBOX *mbox, uint8_t *dataC )
 *
 * Overview:        API Function
 *                  Use this function to write dataC to CAN Bus
 *
 * PreCondition:    None
 *
 * Input:           *mbox, *dataC
 *
 * Output:          return: CAN_OK / CAN_NOK
 *
 * Side Effects:    None
 **********************************************************************/
int8_t CAN_Write( CAN_MAILBOX *mbox, uint8_t *dataC )
{
  int8_t  CAN_Status_Tx = CAN_NOK;
  uint8_t i=0;

  // 50 Fehlversuche entspricht 5,5ms oder CAN_Tx_Voltage von 8 Slaves
  for(i=0;i<CAN_Tx_TimeOut;i++)
  {
    CAN_Status_Tx = ECANSendMessage( mbox->CAN_Id, dataC, mbox->DLC, ECAN_TX_STD_FRAME);
    if(CAN_Status_Tx)   // wenn Tx complete
      break;            // break loop

    Delay100TCYx(11);   // loop time ca 0,1ms => 1 CAN Telegramm bei 1MBaud
  }
  return CAN_Status_Tx;
}



/*********************************************************************
 * Function:        int8_t Load_Mailbox( uint8_t SlaveNo )
 *
 * Overview:        Use this function to initialize all CAN Mailboxes
 *
 * PreCondition:    None
 *
 * Input:           SlaveNo
 *
 * Output:          return: CAN_OK / CAN_NOK
 *
 * Side Effects:    change CAN_Mailbox
 **********************************************************************/
int8_t Load_Mailbox( uint8_t SlaveNo )
{
  gCAN_MAILBOX[0].BOX_No = 0;              // UI --> Master Frame
  gCAN_MAILBOX[0].CAN_Id = 0x2F7;
  gCAN_MAILBOX[0].DLC    = 8;
  gCAN_MAILBOX[0].ACCEPTANCE_FILTER = 0;

  gCAN_MAILBOX[1].BOX_No = 1;              // Master --> UI Frame
  gCAN_MAILBOX[1].CAN_Id = 0x1F0;
  gCAN_MAILBOX[1].DLC    = 4;
  gCAN_MAILBOX[1].ACCEPTANCE_FILTER = 0x1F0;

  gCAN_MAILBOX[2].BOX_No = 2;              // Slave0 --> UI Frame
  gCAN_MAILBOX[2].CAN_Id = 0x206;
  gCAN_MAILBOX[2].DLC    = 8;
  gCAN_MAILBOX[2].ACCEPTANCE_FILTER = 0x206;

  gCAN_MAILBOX[3].BOX_No = 3;              // Slave1 --> UI Frame
  gCAN_MAILBOX[3].CAN_Id = 0x216;
  gCAN_MAILBOX[3].DLC    = 8;
  gCAN_MAILBOX[3].ACCEPTANCE_FILTER = 0x216;

  gCAN_MAILBOX[4].BOX_No = 8;              // Serialcode Frame
  gCAN_MAILBOX[4].CAN_Id = 0x2FA;
  gCAN_MAILBOX[4].DLC    = 8;
  gCAN_MAILBOX[4].ACCEPTANCE_FILTER = 0;


  return CAN_OK;
}



/*********************************************************************
 * Function:        int8_t CAN_Write_UI( void )
 *
 * Overview:        Use this function to write UI Frames to CAN
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          return: CAN_OK / CAN_NOK
 *
 * Side Effects:    None
 **********************************************************************/
int8_t CAN_Write_UI( void )
{
  int8_t CAN_UI_Status = CAN_NOK;

  gCAN_dataC[0] = 0x00;

  gCAN_dataC[0] = (( gCounter << 5 ) | UIMode );        // formating dataC
  gCAN_dataC[1] = ( gVoltage >>   8 );
  gCAN_dataC[2] = ( gVoltage & 0xFF );
  gCAN_dataC[3] = ( gCurrent >>   8 );
  gCAN_dataC[4] = ( gCurrent & 0xFF );
  gCAN_dataC[5] = UI_ERROR;
  gCAN_dataC[6] = 0;
  gCAN_dataC[7] = gCAN_dataC[0] ^ gCAN_dataC[1] ^ gCAN_dataC[2] ^ gCAN_dataC[3] ^ gCAN_dataC[4] ^ gCAN_dataC[5] ^ gCAN_dataC[6] ^ 0x00;

  // send one UI Frame
  CAN_UI_Status = CAN_Write( &gCAN_MAILBOX[0], &gCAN_dataC[0] );

  if(CAN_UI_Status == CAN_OK)
  {
    gCounter++;

    if(gCounter >= MAX_ALIVE)
    {
        gCounter = 0;
    }

  }

  return CAN_UI_Status;
}

/*********************************************************************
 * Function:        int8_t CAN_Read_Master( void )
 *
 * Overview:        Use this function to read Data from Master
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          return: CAN_OK / CAN_NOK
 *
 * Side Effects:    None
 **********************************************************************/
int8_t CAN_Read_Master( void )
{
  uint8_t               data[8];            // Datenpuffer & Pufferlaenge
  uint8_t               DLC = 8;
  uint32_t              id;
  ECAN_RX_MSG_FLAGS     flags;
  int8_t                CAN_Status_Master = CAN_NOK;

  CAN_Status_Master = ECANReceiveMessage( &id, data, &DLC, &flags);
  gCAN_Rx_Overrun = (flags & ECAN_RX_OVERFLOW);

  if(id == 0x1F0)
  {
      CAN_Status_Master = CAN_OK;
      Alive_Master[15] = ( (data[0] & 0xE0) >> 5 );
      UIMode = ( data[0] & 0x1F );

      Check_Alive_Master[15] =  Alive_Master[15];
      Check_Alive_Master[15]++;

     if(Check_Alive_Master[15] >= MAX_ALIVE)
     {
        Check_Alive_Master[15] = 0;
     }
  }

  return CAN_Status_Master;
}

/*********************************************************************
 * Function:        int8_t CAN_Read_Slave( void )
 *
 * Overview:        Use this function to read Data from master and Slaves
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          return: CAN_OK / CAN_NOK
 *
 * Side Effects:    Store Data
 **********************************************************************/
int8_t CAN_Read_Master_Slave( void )
{
  uint8_t               data[8];            // Datenpuffer & Pufferlaenge
  uint8_t               DLC = 8;
  uint32_t              id;
  uint8_t               id_mask;
  ECAN_RX_MSG_FLAGS     flags;
  int8_t                CAN_Status = CAN_NOK;
  uint8_t               i = 0;

  id = 0;
  id_mask = 0;
  
  CAN_Status = ECANReceiveMessage( &id, data, &DLC, &flags);
  gCAN_Rx_Overrun = (flags & ECAN_RX_OVERFLOW);

  PIR5bits.RXBnIF = 0;
  

  if(id == 0x1F0)
  {
      Write_Flag = 1;           // set Flag for Write_UI
  }

   if((id & 0xF00) == 0x100)
   {
       gIRC_MSG_BUF.CAN_ID[0] = id;
       for(i = 0; i < 8; i++)
       {
           gIRC_MSG_BUF.CAN_DATA[0][i] = data[i];
       }

       CAN_Status = CAN_OK;
   }
   else if((id & 0xF00) == 0x200)
   {
       id_mask = id & 0x00F;
       id_mask++;
       gIRC_MSG_BUF.CAN_ID[id_mask] = id;
       for(i = 0; i < 8; i++)
       {
           gIRC_MSG_BUF.CAN_DATA[id_mask][i] = data[i];
       }

       if((id & 0x00F) == 0x006)
       {
           Check_Can_Flag = 1;
       }
       CAN_Status = CAN_OK;
   }

  return CAN_Status;
}

/*********************************************************************
 * Function:        void CAN_Slave_check(uint32_t can_id, uint8_t *ptr)
 *
 * Overview:        Use this function to check CAN Counter
 *                  and store data from Slaves
 *
 * PreCondition:    None
 *
 * Input:           Can ID, Pointer to Can Data
 *
 * Output:          return: CAN_OK / CAN_NOK
 *
 * Side Effects:    None
 **********************************************************************/
void CAN_Slave_check(uint32_t can_id, uint8_t *ptr)
{
    uint8_t     SL_ERROR_flag = 0;
    uint8_t     LTC_TimeOut = 0;
    uint8_t     array_nr_slave;


    if(can_id <= 0x2FF)
    {
        array_nr_slave = (can_id & 0x0F0) +(can_id & 0x00F);
    

        if((can_id & 0x00F) == 0x006)
        {
            Alive_Slave[array_nr_slave] = ( (*(ptr) & 0xE0) >> 5 );
            SL_ERROR_flag = *(ptr+1);

            LTC_TimeOut = (SL_ERROR_flag & 0x10) >> 4;          // Set Errorflag Timeout
            SL_ERROR_flag = SL_ERROR_flag & 0xEF;               // Clear Errorflag Timeout

            if(!(LTC_TimeOut == 0))
            {
                gERROR_Counter++;                     // Counter for Errorflag Timeout
            }
        }
        else
        {
            Alive_Slave[array_nr_slave] = ((*(ptr) & 0xE0) >> 5 );
        }
    }
    
////////////////////////////////////////////////////////////////////////////////
    
    if(Alive_Slave[array_nr_slave] == Check_Alive_Slave[array_nr_slave])
    {
        Check_Alive_Slave[array_nr_slave]++;

    }
    else
    {
        Check_Alive_Slave[array_nr_slave] = Alive_Slave[array_nr_slave];
        Check_Alive_Slave[array_nr_slave]++;

        Count_Error_Slave[array_nr_slave]++;          // Set Count Error


    }

    if(Check_Alive_Slave[array_nr_slave] >= MAX_ALIVE)
    {
        Check_Alive_Slave[array_nr_slave] = 0;
    }

////////////////////////////////////////////////////////////////////////////////
    
    if( gERROR_Counter >= LTC_TimeOut_Max )
    {
        UIMode = ERROR_MODE;
    }

    if( !(SL_ERROR_flag == 0))
    {
        UIMode = ERROR_MODE;
    }

}

/*********************************************************************
 * Function:        void CAN_Master_check(uint32_t can_id, uint8_t *ptr)
 *
 * Overview:        Use this function to check CAN Counter
 *                  and store data from Master
 *
 * PreCondition:    None
 *
 * Input:           Can ID, Pointer to Can Data
 *
 * Output:          None
 *
 * Side Effects:    None
 **********************************************************************/
void CAN_Master_check(uint32_t can_id, uint8_t *ptr)
{
    uint8_t     array_nr_master;
    
    array_nr_master = (can_id & 0x0F0) >> 4;

    if(can_id == 0x1F0)
    {
        Alive_Master[array_nr_master] = ( (*(ptr) & 0xE0) >> 5 );
        UIMode = ( *(ptr) & 0x1F );
    }
    else
    {
        Alive_Master[array_nr_master] = ( (*(ptr) & 0xE0) >> 5 );
    }

////////////////////////////////////////////////////////////////////////////////

    if( Alive_Master[array_nr_master] == Check_Alive_Master[array_nr_master])
    {
        Check_Alive_Master[array_nr_master]++;

    }
    else
    {
        Check_Alive_Master[array_nr_master] =  Alive_Master[array_nr_master];
        Check_Alive_Master[array_nr_master]++;

        Count_Error_Master[array_nr_master]++;          // Set Count Error


    }

    if(Check_Alive_Master[array_nr_master] >= MAX_ALIVE)
    {
        Check_Alive_Master[array_nr_master] = 0;
    }
    

}

/*********************************************************************
 * Function:        void CAN_Alive()
 *
 * Overview:        This function check Error Counter of Slave
 *                  and Master and clear it
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 **********************************************************************/
void CAN_Alive()
{
    uint8_t i;

    for(i=0; i<160; i++)
    {
        if(Count_Error_Slave[i] >= 4)
        {
            UIMode = ERROR_MODE;            // UI Error Case
        }

        Count_Error_Slave[i]    = 0;        // clear Error Slave
    }

    for(i=0; i<16; i++)
    {
        if(Count_Error_Master[i] >= 4)
        {
            UIMode = ERROR_MODE;            // UI Error Case
        }

        Count_Error_Master[i]   = 0;        // clear Error Master
    }
}
/*********************************************************************
 * Function:        int8_t CAN_Write_System( void )
 *
 * Overview:        Use this function to write Serialframe to Master
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          return: CAN_OK / CAN_NOK
 *
 * Side Effects:    None
 **********************************************************************/
int8_t CAN_Write_System( void )
{
  int8_t    CAN_Status_System = CAN_NOK;
  uint16_t  DELAY_t = 0;
  uint16_t  i_time = 0;
  uint8_t   Delay_256us = 4;

    gCAN_Data[0]= gSerial.EE_BYTE.LFT_1;
    gCAN_Data[1]= gSerial.EE_BYTE.LFT_2;
    gCAN_Data[2]= gSerial.EE_BYTE.LIEF;
    gCAN_Data[3]= gSerial.EE_BYTE.BAU_1;
    gCAN_Data[4]= gSerial.EE_BYTE.EE_REV;
    gCAN_Data[5]= gSerial.EE_BYTE.FAM_1;
    gCAN_Data[6]= gSerial.EE_BYTE.GERAET;
    gCAN_Data[7]= gSerial.EE_BYTE.LAEN;

    // LFT1(LSB) and LFT2(MSB) to 16 Bit DELAY_t value
    DELAY_t = gSerial.EE_BYTE.LFT_2;
    DELAY_t = DELAY_t << 8;
    DELAY_t = DELAY_t | gSerial.EE_BYTE.LFT_1;
    // Delay_t 0 - 9999 --> 1 - 10000
    DELAY_t++;

    // Generate delay time with 8 Bit Delay1KTCYx(Delay_256us) function
    i_time = 0;

    while(i_time <= DELAY_t )
    {
        Delay1KTCYx(Delay_256us);
        i_time++;
    }

    CAN_Status_System = CAN_Write( &gCAN_MAILBOX[4], &gCAN_Data[0] );

    i_time = 1;
    i_time = 2;

  return CAN_Status_System;
}

/*********************************************************************
 * Function:        int8_t CAN_Read_System( void )
 *
 * Overview:        Use this function to system initialization
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          return: CAN_OK / CAN_NOK
 *
 * Side Effects:    change ID, readout seriescode
 **********************************************************************/
int8_t CAN_Read_System( void )
{
  uint8_t               data[8];                    // Datenpuffer & Pufferlaenge
  uint8_t               SERNRPRG_check = 0x00;      // XOR Checksum for SERNRPRG
  uint8_t               GETSERNR_check = 0x0E;      // XOR Checksum for GETSERNR
  uint8_t               CHECK_temp = 0;
  int8_t                CAN_Status_System = CAN_NOK;
  uint8_t               DLC = 8;
  uint32_t              id=0;
  ECAN_RX_MSG_FLAGS     flags;
  uint8_t               i;
  uint8_t               DATA_temp[8];

  CAN_Status_System = ECANReceiveMessage( &id, data, &DLC, &flags);
  gCAN_Rx_Overrun = (flags & ECAN_RX_OVERFLOW);

  // Request with ID 0x300 from Master for IDsearch or to programm new Serialcode
  if(id == 0x300)
  {
     // XOR Checksum calculation
     // 'SERNRPRG' => 0x53 ^ 0x45 ^ 0x52 ^ 0x4E ^ 0x52 ^ 0x50 ^ 0x4F ^ 0x47 = 0x00
     // 'GETSERNR' => 0x47 ^ 0x45 ^ 0x54 ^ 0x53 ^ 0x45 ^ 0x52 ^ 0x4E ^ 0x52 = 0x0E
     CHECK_temp = data[0] ^ data[1] ^ data[2] ^ data[3] ^ data[4] ^ data[5] ^ data[6] ^ data[7];

     if(CHECK_temp == SERNRPRG_check)
     {
         STATE_STATUS = SYSTEM_SERNRPRG;
         ECANSetRXM0Value(0xDF6, ECAN_MSG_STD);        // activate 0x308 and 0x309 Frame
     }
     else if(CHECK_temp == GETSERNR_check)
     {
         STATE_STATUS = SYSTEM_GETSERNR;
     }
  }

switch(STATE_STATUS)
        {
            case SYSTEM_SERNRPRG:
                   // Request with ID 0x308 and 0x309 from System to save the new Serialcode
                    if( (id == 0x308) | (id == 0x309))
                    {
                        //Get Data from Master and save to DATA_temp
                        for(i = 0; i < DLC; i++)
                        {
                            DATA_temp[i] = data[i];
                        }
                        if(id == 0x308)
                        {
                             EEwrite(EE_308_a, &DATA_temp[0] );      // Write initialization data to EEPROM
                        }
                        else if(id == 0x309)
                        {
                            EEwrite(EE_309_a, &DATA_temp[0] );      // Write initialization data to EEPROM

                            // Restart main programm
                            UIMode = INITIALISATION_MODE;
                            STATE_STATUS = RUN_UI;
                        }
                    }
            break;
            case SYSTEM_GETSERNR:
                 // Reply with ID 0x2XA from Slave to Master
                CAN_Write_System();

                // Restart main programm
                UIMode = INITIALISATION_MODE;
                STATE_STATUS = RUN_UI;

                break;
            default:
            break;
        }

  return CAN_Status_System;
}